import re
from collections import Callable
from typing import Optional
from datetime import date
from datetime import datetime
from datetime import timedelta

from django.conf import settings

from kinder.core.declaration.models import DSS
from kinder.core.direct.models import DRS
from kinder.core.group.models import Pupil
from kinder.core.helpers import recursive_getattr


CREATE_ORDER_STATUS_MAP = {
    (DSS.RECEIVED, ): (
        110, 'Заявление передано в региональную систему доступности '
             'дошкольного образования.  Заявление зарегистрировано '
             '{declaration__date} с номером {declaration__id}. '
             'Ожидайте рассмотрения в течении 7 дней. '
    ),
    (DSS.ACCEPTED_FOR_CONSIDERING, ): (
        120, 'Начато рассмотрение заявления.'
    ),
    (DSS.DUL_CONFIRMATING, DSS.MED_CONFIRMATING,
     DSS.TUTOR_CONFIRMATING, DSS.ZAGS_CHECKING): (
        130, 'Для подтверждения данных вам необходимо представить в '
             '{declaration__mo__address_full} в срок {auto_change_days} '
             'следующие документы: {doc_comment_format}'
    ),
    (DSS.REFUSED, ): (
        150, 'Вам отказано в предоставлении услуги по текущему заявлению по '
             'причине {comment}. '
             'Вам необходимо обратиться по адресу: '
             '{declaration__mo__address_full}.'
    ),
    (DSS.ACCEPTED, ): (
        250, 'Ваш ребенок зачислен в '
             '{last_pupil__grup__unit__name}, расположенную по '
             'адресу {last_pupil__grup__unit__address_full}. '
             'На основании договора от {last_pupil__date_in_order_to_doo}.'
    )
}


APPLICATION_REVIEWED = (
    140,
    'Ваше заявление рассмотрено. Индивидуальный номер заявления '
    '{declaration__id}. Ожидайте направления в выбранную образовательную '
    'организацию после {declaration__desired_date}.'
)

AWAITING_DIRECTION = (
    160,
    'В настоящее время в образовательных организациях, указанных в '
    'заявлении, нет свободных мест, соответствующих запрашиваемым в '
    'заявлении условиях. '
)

DEFAULT_STATUS = (
    130,
    'Необходимо подтвердить данные заявления.'
)

# После отправки в ЭДС запроса ApplicationRequest с изменениями заявления
DECLARATION_CHANGED = 1
# Вкладка "Изменения с ЕПГУ" - выбрать запись - Отказать
DECLARATION_CHANGED_REFUSED = 2

# Относятся к смене статуса направления (при этом должен менятся статус заявки)
# Направлен в дошкольную образовательную организацию
DIRECTED_TO_DOO = 3
# Потребность в получении места не подтверждена
NEED_NOT_CONFIRMED = 4
# Заявитель отказался от предоставленного места
DELEGATE_REFUSED = 5
# Формирование заявления о приеме
DECLARATION_FORMATION = 6
# Ожидание заключения договора
WAITING_FOR_CONTRACT = 7

UPDATE_ORDER_STATUS_MAP = {
    (DECLARATION_CHANGED,): (
        170,
        'В заявления для направления были внесены следующие изменения: '
        '{get_declaration_changes}.'
    ),
    (DECLARATION_CHANGED_REFUSED,): (
        180,
        'Вам отказано в изменении заявления по причине: {get_refused_reason}.'
    ),
    (DIRECTED_TO_DOO,): (
        190,
        'Вам предоставлено место в {direct__group__unit__name} в '
        'группу {direct__group__name} ({direct__group__sub_age_cat__name}) в '
        'соответствии с {direct__document_details}. '
        'Вам необходимо явиться по адресу {direct__group__unit__address_full}'
        ' {get_direct_reject_date}.'
    ),
    (NEED_NOT_CONFIRMED,): (
        210,
        'Действие Вашего заявления приостановлено по причине '
        '{didnt_come_comment_format}.'
    ),
    (DELEGATE_REFUSED,): (
        220,
        'Действия по заявлению приостановлены по причине Вашего отказа от '
        'предоставленного места. Вам необходимо изменить заявление '
        'либо отозвать его.'
    ),
    (DECLARATION_FORMATION,): (
        230,
        'Согласие с предоставленным местом направленно на рассмотрение в '
        '{declaration__mo__name}.'
    ),
    (WAITING_FOR_CONTRACT,): (
        240,
        'Ваше заявление рассмотрено. Вам необходимо заключить договор до '
        '{get_date_for_contract}.'
    )
}


class CreateOrderStatusMapper:
    """
    Маппинг статусов для CreateOrderRequest, в основе используется маппинг
    CREATE_ORDER_STATUS_MAP. Возвращает код и комментарий для ЕПГУ.

    Подготовливает комментарий, заменяя часть сообщения на значения
    атрибутов класса или на результаты вызова функций класса.

    Описание https://conf.bars.group/pages/viewpage.action?pageId=123901022
    """

    def __init__(self, declaration, direct):
        self.declaration = declaration
        self.direct = direct
        self.status_map = CREATE_ORDER_STATUS_MAP

    @property
    def last_pupil(self) -> Optional[Pupil]:
        """Последнее зачисление ребёнка (фактическое/плановое/None)

        Не используется информация о направлении, поскольку чаще всего оно
        не указано явно выше в коде и берётся первое попавшееся направление
        ребёнка. Это может привести к неправильному комментарию о зачислении.

        Данный вариант также не идеален, но хотя бы лучше выдать информацию
        о существующем зачислении, чем о зачислении, которого на самом деле нет.
        """
        pupil = Pupil.fact_pupils.for_child(
            self.declaration.children_id).order_by('-id').first()
        if not pupil:
            pupil = Pupil.plan_pupils.for_child(
                self.declaration.children_id).order_by('-id').first()
        return pupil

    @staticmethod
    def doc_comment_format(status_data):
        """Возвращает тип документа для комментария"""

        status_code = status_data['status__code']

        if status_code == DSS.DUL_CONFIRMATING:
            return (
                'Свидетельство о рождении, Документ, подтверждающий прописку.'
            )
        elif status_code == DSS.MED_CONFIRMATING:
            return 'Заключение ПМПК.'
        elif status_code == DSS.TUTOR_CONFIRMATING:
            return 'Документы, подтверждающие опеку.'
        elif status_code == DSS.ZAGS_CHECKING:
            return 'Свидетельство о рождении.'
        else:
            return ''

    def check_is_application_reviewed(self, status_code):
        """Проверка для статуса концентратора "Заявление рассмотрено"."""

        application_reviewed = False

        alloc_date_bgn = self.declaration.mo.alloc_date_bgn
        alloc_date_end = self.declaration.mo.alloc_date_end
        alloc_calc_date = self.declaration.mo.alloc_calc_date
        today = date.today()

        check_statuses = [
            DSS.REGISTERED, DSS.WANT_CHANGE_DOU, DSS.PRIV_CONFIRMATING
        ]

        if alloc_date_bgn and alloc_date_end:
            if alloc_date_bgn <= today <= alloc_date_end:
                if (alloc_calc_date and
                        self.declaration.desired_date > alloc_calc_date and
                        status_code in check_statuses):
                    application_reviewed = True

            elif (self.declaration.desired_date > today and
                    status_code in check_statuses):
                application_reviewed = True

        return application_reviewed

    def check_is_awaiting_direction(self, status_code):
        """Проверка для статуса концентратора "Ожидает направления"."""

        awaiting_direction = False

        alloc_date_bgn = self.declaration.mo.alloc_date_bgn
        alloc_date_end = self.declaration.mo.alloc_date_end
        alloc_calc_date = self.declaration.mo.alloc_calc_date
        today = date.today()

        check_statuses = [
            DSS.REGISTERED, DSS.WANT_CHANGE_DOU, DSS.PRIV_CONFIRMATING
        ]

        if alloc_date_bgn and alloc_date_end:
            if alloc_date_bgn <= today <= alloc_date_end:
                if (alloc_calc_date and
                        self.declaration.desired_date <= alloc_calc_date and
                        status_code in check_statuses):
                    awaiting_direction = True

            elif (self.declaration.desired_date <= today and
                    status_code in check_statuses):
                awaiting_direction = True

        return awaiting_direction

    def prepare_comment(self, status_data, epgu_comment):
        """
        Подготовка комментария, заменяет часть сообщения на значения
        атрибутов класса или на результат вызова функций класса.

        """

        param_regex = r'({(\S+)})'

        for attr in re.findall(param_regex, epgu_comment):
            comment_attr = recursive_getattr(self, attr[1]) or ' '

            if isinstance(comment_attr, Callable):
                epgu_comment = re.sub(
                    attr[0], comment_attr(status_data) or '', epgu_comment)
            else:
                if isinstance(comment_attr, date):
                    comment_attr = comment_attr.strftime(settings.DATE_FORMAT)
                elif isinstance(comment_attr, datetime):
                    comment_attr = comment_attr.strftime(
                        f'{settings.DATE_FORMAT} %H:%M:%S')

                epgu_comment = re.sub(attr[0], str(comment_attr), epgu_comment)

        return epgu_comment

    def __getitem__(self, status_data):
        status_code = status_data['status__code']

        # Комментарий к смене статуса
        self.comment = status_data.get('comment', '')
        self.auto_change_days = status_data.get('auto_change_days', '')

        for statuses_keys in self.status_map.keys():
            if status_code in statuses_keys:
                epgu_status, epgu_comment = self.status_map.get(statuses_keys)

                return epgu_status, self.prepare_comment(
                    status_data, epgu_comment)

        if self.check_is_application_reviewed(status_code):
            code, comment = APPLICATION_REVIEWED
            return code, self.prepare_comment(status_data, comment)
        elif self.check_is_awaiting_direction(status_code):
            return AWAITING_DIRECTION
        else:
            return DEFAULT_STATUS


class UpdateOrderStatusMapper(CreateOrderStatusMapper):
    """Маппинг статусов для UpdateOrderRequest"""

    def __init__(
            self, declaration, direct, declaration_status_changed=False,
            direct_status_log=None, event=None, declaration_changes_rows=None,
            reject_changes_comment=None
    ):

        super().__init__(declaration, direct)

        self.declaration_status_changed = declaration_status_changed
        self.direct_status_log = direct_status_log
        self.event = event
        self.declaration_changes_rows = declaration_changes_rows
        self.reject_changes_comment = reject_changes_comment

        self.status_map.update(UPDATE_ORDER_STATUS_MAP)

    def prepare_comment(self, status_data, epgu_comment):
        """
        При изменении полей заявления передавать только последний статус и
        фиксированный комментарий
        "Заявление было изменено во время личного приема."

        """

        if (not self.declaration_status_changed and not self.direct_status_log
                and not self.event):
            return 'Заявление было изменено во время личного приема.'
        else:
            return super().prepare_comment(status_data, epgu_comment)

    def get_declaration_changes(self, status_data):
        if self.declaration_changes_rows:
            from concentrator.smev3_v321.application_request.application import (
                changes_to_str)

            return changes_to_str(self.declaration_changes_rows)

    def get_refused_reason(self, status_data):
        if self.reject_changes_comment:
            return self.reject_changes_comment

    def get_direct_reject_date(self, status_data):
        if (self.direct.days_for_reject_direct and
                self.direct_status_log['created_at']):

            end_date = (
                self.direct_status_log['created_at'] +
                timedelta(days=self.direct.days_for_reject_direct)
            )

            return f'до {end_date.strftime("%d.%m.%Y")}'

    def didnt_come_comment_format(self, status_data):
        """
        Для статуса "Оказано либо отказался" берем текст из поля '
        'Причина. Для статуса "Не явился" пишем: "окончания срока '
        'действия направления [если статус был Направлен в ДОО, '
        'указываем значение Дата смены статуса на '
        '"Направлен в ДОО" + значение из поля "Срок действия направления '
        'в Направлен в ДОО" в справочнике МО, вкладка Комплектование; если '
        'статус Заключение договора, указываем значение Дата смены статуса '
        'на "Заключение договора" + значение из поля '
        '"Срок заключения договора для направления" в справочнике МО, '
        'вкладка Комплектование].

        """

        status_code = status_data.get('status__code')
        created_at = status_data.get('created_at')
        reason = status_data.get('reason', '')
        old_status = status_data.get('old_status')

        if status_code == DRS.REFUSED:
            return reason
        elif status_code == DRS.REJECT:
            comment_str = 'окончания срока действия направления '

            if old_status == DRS.REGISTER:
                end_date = created_at + timedelta(
                    days=self.declaration.mo.days_for_reject_direct)

                return comment_str + end_date.strftime(settings.DATE_FORMAT)
            elif old_status == DRS.DOGOVOR:
                end_date = created_at + timedelta(
                    days=self.declaration.mo.days_for_direct_dogovor)

                return comment_str + end_date.strftime(settings.DATE_FORMAT)

    def get_date_for_contract(self, status_data):
        """
        Дата смены статуса на Заключение договора + "Срок заключения договора
        для направлений" (если заполнено) в МО,
        которое указано в заявлении.

        """

        created_at = status_data.get('created_at')
        end_date = created_at + timedelta(
            days=self.declaration.mo.days_for_direct_dogovor)

        return end_date.strftime(settings.DATE_FORMAT)


# Наименования по умолчанию типов ВС и кодов методов сервисов семейства
# OrderRequest
DEFAULT_ORDER_REQUEST_MESSAGE_TYPE = 'OrderRequest'
DEFAULT_CREATE_ORDER_REQUEST_METHOD_NAME = 'CreateOrderRequest'
DEFAULT_UPDATE_ORDER_REQUEST_METHOD_NAME = 'UpdateOrderRequest'
